#!/usr/bin/env python
# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

"""imports <model>, loops through its SQLObjects and
outputs new class defs using columnsFromSchema()
TODO: support model types other than SQLObject"""

##
## NOTE: this script will probably move soon.

import optparse
import os
import sys
import sre
try:
    from sqlobject.declarative import DeclarativeMeta
except ImportError:
    # pre 0.7 ...
    from sqlobject.main import MetaSQLObject as DeclarativeMeta
import sqlobject

class SO_Info(object):
    _from_name = None
    _so = None
    cols = []
    klass = None
    style = None
    table = None
    def __init__(self, name, so=None):
        self.cols = []
        self._from_name = name
        self._so = so
        try:
            so.sqlmeta
        except AttributeError:
            # pre 0.7 ...
            self.klass = name # hmm
            self.style = so._style
            self.table = so._table
        else:
            self.klass = so.sqlmeta.soClass.__name__
            self.style = so.sqlmeta.style
            self.table = so.sqlmeta.table
        #print 'name=', self._from_name
        #print 'table=', self.table

class SOCol_Info(object):
    klass = None
    name = None
    kw = []
    def __init__(self, klass=None, name=None, kw=[]):
        self.klass = klass
        self.name = name
        self.kw = kw
    
    def __repr__(self):
        return '%s = %s(%s)' % (self.name, 
                                self.klass,
                                ', '.join(['%s=%s' % (k,v) for k,v in self.kw]))

class FKey_Info(object):
    table = None
    def __init__(self, table=None):
        self.table = table
    
    def dbTableToPythonClass(self, table, style):
        """translates a discovered foreign key 
        table name into a SQLObject class
        """
        return style.dbTableToPythonClass(table)
    dbTableToPythonClass = classmethod(dbTableToPythonClass)

def build_fkey_map(model):
    map = {}
    for so_info in iter_SO(model):
        map[so_info.table] = so_info.klass
    return map

def iter_SO(model, sort=False):
    names = dir(model)
    if sort:
        names.sort()
    for name in names:
        obj = getattr(model, name)
        if type(obj) is DeclarativeMeta and name not in (
                        'SQLObject', 'sqlmeta', 'OneToMany', 'ManyToMany'):
            
            yield SO_Info(name, so=obj)
    else:
        raise StopIteration

def build_model_map(model, fkey_map=None, fkey_trans=None):    
    isnum = sre.compile(r'^[0-9\.]+$').match
    so_map = []
    
    if fkey_trans is None:
        fkey_trans = FKey_Info.dbTableToPythonClass
        
    for so_info in iter_SO(model, sort=True):
        so = so_info._so
        
        if so_info.style.longID:
            skipID = so_info.style.idForTable(so_info.table)
            skipID = so_info.style.dbColumnToPythonAttr(skipID)
        else:
            skipID = None # cuz `id' is already skipped
        
        for col in so._connection.columnsFromSchema(so_info.table, so):
            if skipID and col.name == skipID: 
                continue
            kw = []
            try:
                col_kw = col._kw
            except AttributeError:
                # pre 0.7 ...
                col_kw = col.kw
                
            for_key = None
            for k,v in col_kw.items():
                if k == 'foreignKey':
                    for_key = FKey_Info(table = v) 
                if k in ('dbName','foreignKey'):
                    continue
                if issubclass(v.__class__, str) and not isnum(v):
                    v = "'%s'" % v.replace("'","\'")
                kw.append( (k,v) )
            
            col_class = for_key and 'ForeignKey' or col.__class__.__name__
            col_name = col.name
            if for_key:
                # probably a much better way to do this but I'm too
                # lazy to poke around sqlobject source
                if col_name.endswith('ID'):
                    col_name = col_name[0:-2]
                
                assert fkey_map is not None, \
                        "a ForeignKey col was detected but you " \
                        "did not provide an fkey_map to lookup tables"
                try:
                    fkey_ref = fkey_map[for_key.table]
                except KeyError:
                    fkey_ref = fkey_trans(for_key.table, so_info.style)
                kw.append( ('foreignKey', "'%s'" % fkey_ref ) )
            
            so_info.cols.append( SOCol_Info(klass=col_class, name=col_name, kw=kw) )
        so_map.append(so_info)
            
    return so_map

class SO_CodeMerger(object):
    '''merges the generated SQLObject defs (so_map) into the 
    pre-existing code found in the file-like object passed into merge()
    
    FIXME: descending into a class isn't very smart.  I think if you had:
    \class TableObject:
        \class sqlmeta:
            attr = val
    where attr was the same name as a column in your DB, weird stuff
    will happen
    
    '''
    
    klass_p = sre.compile(r'^(?P<indent>\s*)class (?P<name>[a-zA-Z0-9_]+).*$')
    attr_p = sre.compile(r'^(?P<indent>\s*)(?P<name>[a-zA-Z0-9_]+)\s*=.*$')
    so_map = {}
    so_pool = []
    exist_code = None
    line = None
    clobber = False
    
    def __init__(self, so_map):
        self.so_map = so_map
        
        self.so_lookup = {}
        for so in self.so_map:
            self.so_lookup[so.klass] = so
        self.so_pool = self.so_lookup.keys()
        
    def next(self):
        self.line = self.exist_code.readline()
        if self.line=='':
            # EOF
            raise StopIteration

    def ignore_line(self, line):
        # just echo ..
        print line,

    def append_dyn_cols(self, col_map):
        added_some = False
        for name, col in col_map.items():
            added_some = True
            print '    ' + repr(col)
            
        if added_some:
            print '' # make up for the lost space

    def descend_class(self, klsinfo):
        kls = klsinfo.group('name')
    
        if kls not in self.so_pool:
            self.ignore_line(self.line)
            return
        so_info = self.so_lookup[kls]
    
        # print orig. class def :
        self.ignore_line(self.line)
    
        self.next()
        col_map = dict([ (c.name, c) for c in so_info.cols ])
    
        # dummy indent == one level further ...
        while self.line.startswith(klsinfo.group('indent') + ' ') \
                                                or self.line.strip()=='':
        
            # compare columns ...
        
            colm = self.attr_p.match(self.line)
        
            if colm and col_map.has_key(colm.group('name')):
                c = colm.group('name')
                if self.clobber:
                    print '    ' + repr( col_map[c] )
                else:
                    self.ignore_line(self.line)
                col_map.pop(c)
            else:
                self.ignore_line(self.line)
            try:
                self.next()
            except StopIteration:
                self.append_dyn_cols(col_map)
                raise
    
        # indent level changed but no EOF yet :
        self.append_dyn_cols(col_map)
    
        klsinfo = self.klass_p.match(self.line)
        if klsinfo:
            self.descend_class(klsinfo)
        else:
            self.ignore_line(self.line) 
            
    def merge(self, exist_code, clobber=False):
        
        self.exist_code = exist_code
        self.clobber = clobber
        self.line = None
        
        try:
            while 1:
                self.next()
    
                klsinfo = self.klass_p.match(self.line)
                if klsinfo:
                    self.descend_class(klsinfo)
                else:
                    self.ignore_line(self.line)
        except StopIteration:
            pass

def to_code(so_map):
    
    print ''
    print 'from sqlobject import *'
    print ''
    
    for so_info in so_map:
        print 'class ' + so_info.klass + '(SQLObject):'
        for col in so_info.cols:
            print '    ' + repr(col)
        print ''

def main(inp=sys.argv[1:], fkey_trans=None):
    parser = optparse.OptionParser(
        usage='%prog package.model' + "\n\n" + __doc__)
    parser.add_option('-m','--merge', action='store_true',
        help="merge new class defs with existing defs")
    parser.add_option('-M','--merge_file', action='store',
        help=   "merge new class defs with existing defs in this file "
                "(when it's too hard to guess the code file)")
    parser.add_option('--clobber', action='store_true',
        help=   "use a db-discovered column even if it "
                "already exists in the code")
    parser.add_option('--dsn', help="sets connection to DSN")
    (options,args) = parser.parse_args()
    
    name, finalmod = os.path.splitext(args[0])
    if not finalmod:
        mod = __import__(name, globals(), locals(), [])
    else:
        if finalmod.startswith('.'):
            finalmod = finalmod[1:]
        mod = __import__(name, globals(), locals(), [finalmod]) 
        mod = getattr(mod, finalmod)
    
    if options.dsn:
        sqlobject.sqlhub.processConnection = \
                                sqlobject.connectionForURI(options.dsn)
        
    fkey_map = build_fkey_map(mod)
    so_map = build_model_map(mod, fkey_map=fkey_map, fkey_trans=fkey_trans)
    
    if options.merge or options.merge_file:
        if options.merge_file is None:
            fname = mod.__file__
            if not fname.endswith('.py'):
                # probably compiled ...
                base, ext = os.path.splitext(os.path.basename(fname))
                fname = os.path.join(os.path.dirname(fname), base + '.py')
        else:
            fname = options.merge_file
            
        m = SO_CodeMerger(so_map)
        m.merge(open(fname, 'r'), clobber=options.clobber)
    else:
        to_code(so_map)

# OUCH! refactor me (code in genfix and genmodel)
# prepare for sqlobject monkey patch :( ...
# this is so that foreign key lookups work right when 
# there are multiple schemas having the same table 
# (perfectly legal, but sqlobject only finds the first primary key)
import re
def columnsFromSchema(self, tableName, soClass):

    keyQuery = """
    SELECT pg_catalog.pg_get_constraintdef(oid) as condef
    FROM pg_catalog.pg_constraint r
    WHERE r.conrelid = %s::regclass AND r.contype = 'f'"""

    colQuery = """
    SELECT a.attname,
    pg_catalog.format_type(a.atttypid, a.atttypmod), a.attnotnull,
    (SELECT substring(d.adsrc for 128) FROM pg_catalog.pg_attrdef d
    WHERE d.adrelid=a.attrelid AND d.adnum = a.attnum)
    FROM pg_catalog.pg_attribute a
    WHERE a.attrelid =%s::regclass
    AND a.attnum > 0 AND NOT a.attisdropped
    ORDER BY a.attnum"""
    
    # kumar: add limit 1 to get primary key for first rel in schema search path
    primaryKeyQuery = """
    SELECT pg_index.indisprimary,
        pg_catalog.pg_get_indexdef(pg_index.indexrelid)
    FROM pg_catalog.pg_class c, pg_catalog.pg_class c2,
        pg_catalog.pg_index AS pg_index
    WHERE c.relname = %s
        AND c.oid = pg_index.indrelid
        AND pg_index.indexrelid = c2.oid
        AND pg_index.indisprimary
    LIMIT 1
    """

    keyData = self.queryAll(keyQuery % self.sqlrepr(tableName))
    keyRE = re.compile(r"\((.+)\) REFERENCES (.+)\(")
    keymap = {}

    for (condef,) in keyData:
        match = keyRE.search(condef)
        if match:
            field, reftable = match.groups()
            keymap[field] = reftable.capitalize()

    primaryData = self.queryAll(primaryKeyQuery % self.sqlrepr(tableName))
    primaryRE = re.compile(r'CREATE .*? USING .* \((.+?)\)')
    primaryKey = None
    for isPrimary, indexDef in primaryData:
        match = primaryRE.search(indexDef)
        assert match, "Unparseable contraint definition: %r" % indexDef
        assert primaryKey is None, "Already found primary key (%r), then found: %r" % (primaryKey, indexDef)
        primaryKey = match.group(1)
    assert primaryKey, "No primary key found in table %r" % tableName
    if primaryKey.startswith('"'):
        assert primaryKey.endswith('"')
        primaryKey = primaryKey[1:-1]

    colData = self.queryAll(colQuery % self.sqlrepr(tableName))
    results = []
    if self.unicodeCols:
        client_encoding = self.queryOne("SHOW client_encoding")[0]
    for field, t, notnull, defaultstr in colData:
        if field == primaryKey:
            continue
        colClass, kw = self.guessClass(t)
        if self.unicodeCols and colClass == col.StringCol:
            colClass = col.UnicodeCol
            kw['dbEncoding'] = client_encoding
        kw['name'] = soClass.sqlmeta.style.dbColumnToPythonAttr(field)
        kw['dbName'] = field
        kw['notNone'] = notnull
        if defaultstr is not None:
            kw['default'] = self.defaultFromSchema(colClass, defaultstr)
        elif not notnull:
            kw['default'] = None
        if keymap.has_key(field):
            kw['foreignKey'] = keymap[field]
        results.append(colClass(**kw))
    return results
from sqlobject.postgres import pgconnection
pgconnection.PostgresConnection.columnsFromSchema = columnsFromSchema
    
if __name__ == '__main__':
    main()
