
import os
from os import path
import unittest
import sys
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from StringIO import StringIO as StringIOInterface

import logging
import testtools
from testtools import bgtests
from testtools.bgtests import Reporter, TestablePackage, PackageTestHandler, SvnInspector
from testtools.io import TempIO
from subprocess import call, STDOUT
import SimpleXMLRPCServer
import socket
from testtools.stubs import mkinterface
import xmlrpclib


# resource denied :
# from test.test_socketserver import pickport
seed = 0
def pickport():
    global seed
    seed += 1
    return 10000 + (os.getpid() % 1000)*10 + seed

class stub_smtplib:
    class SMTP:
        sent = False
        messages = []
        def __init__(self,*a,**kw): pass
        def sendmail(self, em_from, notify, msg):
            self.__class__.sent = True
            self.__class__.messages.append(msg)
        def quit(self): pass

class BahTests(TestablePackage):
    ex = 'nosetests -v'
    svnpath = 'file://%s/bah' # NOTE! %s is just for testing
    mock = {} # mock vars
    
    def recognizes_file(self, filepath, code=None):
        return filepath.startswith('bah/')
    
    def setup(self):
        self.__class__.mock['was_setup'] = True
        self.__class__.mock['listdir'] = os.listdir('.')

class HumbugTests(TestablePackage):
    ex = 'nosetests -v'
    svnpath = 'file://%s/humbug' # NOTE! %s is just for testing
    mock = {}
    
    def recognizes_file(self, filepath, code=None):
        return filepath.startswith('humbug/')
    
    def setup(self):
        self.__class__.mock['was_setup'] = True
        self.__class__.mock['listdir'] = os.listdir('.')

class stub_req(mkinterface(socket.socket)):
    """just like how a running SimpleXMLRPCServer would receive an incoming request. 
    
    """
    def __init__(self, args):
        self._s_args = args
        content = xmlrpclib.dumps( self._s_args, methodname='trigger_from_svnlook')
        header = "\r\n".join((  "POST /RPC2 HTTP/1.0", 
                                "Host: localhost:9999", 
                                "User-Agent: xmlrpclib.py/1.0.1 (by www.pythonware.com)", 
                                "Content-Type: text/xml", 
                                "Content-Length: %s" % len(content), ))
                            
        self._s_request = StringIO( "\r\n".join((header,content)) )
        self._s_response = StringIO()
    
    def makefile(self,*a,**kw):
        
        class SocketBuffer(mkinterface(StringIOInterface)): 
            closed = self._s_request.closed
            def read(proxy, *a,**kw):
                return self._s_request.read(*a,**kw)
            def readline(proxy, *a,**kw):
                return self._s_request.readline(*a,**kw)
            def write(proxy, *a,**kw):
                return self._s_response.write(*a,**kw)
                
        return SocketBuffer()

def test_export():
    assert 'bgtests' in testtools.__all__

def test_stream_capture():
    """assert Reporter can capture its stream for emailing"""
    
    strout = StringIO()
    strout.write('possible pre-logging that should not be captured')
    
    out = Reporter( stream=strout, notify='bah@bah-humbug.org' )
    print >>out, 'some info'
    print >>out, 'some more info'
    
    buffed = out.buffer.getvalue().replace("\n","")
    assert buffed == "some infosome more info"
    
    # make sure we have separate buffers :
    out2 = Reporter( stream=strout, notify='bah@bah-humbug.org' )
    print >>out2, 'humbug'
    assert out2.buffer.getvalue().strip() == 'humbug'    
            
class TestsWithRepos(unittest.TestCase):
    def setUp(self):
        """sets up a tmp repository and commits a few things.
        
        also hooks onto bgtests' smtplib so that actual emails 
        aren't sent but captured instead.
        
        """
        self.cwd = os.getcwd()
        self.real_smtplib = bgtests.smtplib
        bgtests.smtplib = stub_smtplib
        stub_smtplib.SMTP.sent = False
        stub_smtplib.SMTP.messages = []
        
        self.tmp = TempIO()
        self.co_tmp = TempIO()
        
        self.repo = self.tmp.mkdir('repo')
        self.null = file('/dev/null','w') # StringIO no fileno()
        call( ['svnadmin', 'create', self.repo], stdout=self.null )
        
        self.co_tmp.putfile('bah/test_bah.py', """def test_bah(): assert False""")
        call(   [   'svn', 'import', '-m', 'the bah test',  
                    path.join(self.co_tmp.root, 'bah'), 
                    'file://' + path.join(self.repo, 'bah')],
                stdout=self.null )
                
        self.co_tmp.putfile('humbug/test_humbug.py', """def test_humbug(): raise RuntimeError""")
        call(   [   'svn', 'import', '-m', 'the humbug test',  
                    path.join(self.co_tmp.root, 'humbug'), 
                    'file://' + path.join(self.repo, 'humbug')],
                stdout=self.null )
                
        BahTests.mock = {}
        HumbugTests.mock = {}
    
    def tearDown(self):
        bgtests.smtplib = self.real_smtplib
        del self.tmp
        del self.co_tmp
        os.chdir(self.cwd)
    
    def test_package_server(self):        
        
        hum = HumbugTests()
        hum.svnpath = hum.svnpath % self.repo
        
        info =  "blah blah, last change by kumar \n'I broke lots of stuff in trunk'\n"
        changed = "U   humbug/decoy.txt"
        
        def assert_package_tested(buf):
            buffed = buf.getvalue()
            assert 'RuntimeError' in buffed, "unexpected: %s" % buffed
        
        buf = StringIO()
        handler = PackageTestHandler([hum], stream=buf)
        
        # first a sanity check without xmlrpc ...
        handler.trigger_from_svnlook(info, changed, '2')
        assert_package_tested(buf)
        
        buf = StringIO()
        handler = PackageTestHandler([hum], stream=buf)
        
        serv = SimpleXMLRPCServer.SimpleXMLRPCServer(("localhost", pickport()))
        serv.register_introspection_functions()
        serv.register_instance(handler)
        
        req = stub_req( (info, changed, '2') )
        serv.process_request( req, '127.0.0.1')
        
        assert_package_tested(buf)
        # expect False because tests failed ...
        assert '<value><boolean>0</boolean></value>' in req._s_response.getvalue()
    
    def test_SvnInspector(self):
        meta = SvnInspector(self.repo)
        
        info = meta.info()
        assert 'kumar' in info
        assert 'the humbug test' in info
        
        changed = [ln for ln in meta.changed()]
        assert "A   humbug/test_humbug.py\n" in changed, "unexpected: %s" % changed
        
        meta = SvnInspector(self.repo, revision=1)
        changed = [ln for ln in meta.changed()]
        assert "A   bah/test_bah.py\n" in changed, "unexpected: %s" % changed

    def test_reporter_callback(self):
        buf = StringIO()
        
        def skip_notify(buffer):
            assert buffer == buf.getvalue()
            return True
        
        out = Reporter( stream=buf, notify='bah@bah-humbug.org', skip_notify=skip_notify )
        assert not stub_smtplib.SMTP.sent
        
    def test_run(self):        
    
        buf = StringIO()
        r = bgtests.run('file://%s/bah' % self.repo, notify='a@a', test_cmd='nosetests -v', stream=buf)
        assert r is False
        assert stub_smtplib.SMTP.sent
        assert 'AssertionError' in stub_smtplib.SMTP.messages[0]
        buffed = buf.getvalue()
        assert 'AssertionError' in buffed, "not found in : '%s'" % buffed
    
    def test_run_packages(self):
        
        # commit a change that affects all so that 
        # both suites will run ...
        self.co_tmp.newdir('work')
        self.co_tmp.checkout = self.co_tmp.mkdir('work/co')
        os.chdir(self.co_tmp.checkout)
        
        call(   ['svn', 'co', 'file://%s' % self.repo, '.' ],
                stdout=self.null )
        
        # make sure we are running tests in the correct dir:
        self.co_tmp.putfile('work/in_correct_dir.py', "#nop")
        self.co_tmp.putfile('work/co/bah/test_in_correct_dir.py', 
                            """def test(): 
                                import sys
                                sys.path.insert(0, '..')
                                import in_correct_dir
                            """)
                            
        call(   ['svn', 'add', 'bah/test_in_correct_dir.py' ], stdout=self.null )
        
        # just some unique exceptions to look for :
        self.co_tmp.putfile('work/co/bah/test_bah.py', 
                            """def test(): raise AttributeError""")
        self.co_tmp.putfile('work/co/humbug/test_humbug.py', 
                            "\n".join(['print "OUTPUT"', 'def test(): raise ValueError']))
        
        call(   ['svn', 'ci', '-m', 'tweaked bah and humbug tests' ],
                stdout=self.null )
        
        buf = StringIO()
        bah = BahTests()
        hum = HumbugTests()
        for o in (bah,hum):
            o.svnpath = o.svnpath % self.repo
            
        r = bgtests.run_packages( [bah,hum], repos_root=self.repo, 
                                    stream=buf, notify='a@a', tmp_dir=self.co_tmp.work)
        assert r is False
        
        buffed = buf.getvalue()
        # assert the user info and stuff is there (header)
        assert os.environ['USER'] in stub_smtplib.SMTP.messages[0]
        assert 'AttributeError' in stub_smtplib.SMTP.messages[0]
        assert 'test_in_correct_dir' in stub_smtplib.SMTP.messages[0]
        assert 'ImportError' not in stub_smtplib.SMTP.messages[0]
        
        msg = stub_smtplib.SMTP.messages[1]
        assert os.environ['USER'] in msg
        assert 'ValueError' in msg
        # cuz we want the summary at the bottom ...
        self.assertNotEqual( 'OUTPUT', msg.splitlines()[-1])
        
        # assert cleanup :
        for f in os.listdir(self.co_tmp.work):
            assert not (f.startswith('BahTests') or f.startswith('HumbugTests'))
        
        # do the same thing but mocked with a stream ...
        
        class repos_meta:
            changed = StringIO("U   unknown/file.txt").readlines
            info = StringIO("blah blah, last change by kumar \n'"
                            "I broke lots of stuff in trunk'\n").getvalue
                
        buf = StringIO()
        bgtests.run_packages( [ HumbugTests()], repos_root=self.repo, 
                                stream=buf, repos_meta=repos_meta)
        buffed = buf.getvalue()
        assert 'ValueError' not in buffed
        
    def test_TestablePackage(self):
        
        bah = BahTests()
        bah.svnpath = bah.svnpath % self.repo
        buf = StringIO()
        r = bah.runtests( stream=buf, notify=None )
        assert r is False
        
        assert bah.recognizes_file('bah/test_bah.py')
        assert bah.recognizes_file('bah/something/something')
        assert not bah.recognizes_file('booby/test_booby.py')
        
        assert BahTests.mock['was_setup']
        # assert we are actually in the root of the package during setup :
        assert 'test_bah.py' in BahTests.mock['listdir']
        assert not stub_smtplib.SMTP.sent
        buffed = buf.getvalue()
        assert 'AssertionError' in buffed, "not found in : '%s'" % buffed
        