
import testtools
from testtools.stubs import *
import types

def Humbert():
    return 'HUMBERT'
def Delores():
    return 'DELORES'
def Lolita():
    return 'LOLITA'

def test_export():
    assert 'stubs' in testtools.__all__
    
class TestStub:
    def assert_humbert_restorable(self):
        global Humbert
        Humbert = Humbert.restore()
        assert Humbert() == 'HUMBERT'
        assert id(Humbert) == self.orig_id
        
    def assert_call_humbert_stops(self):
        global Humbert
        try:
            Humbert()
        except StopIteration:
            pass
        else:
            raise AssertionError, 'expected to catch StopIteration'
        
    def setUp(self):
        global Humbert
        self.orig_id = id(Humbert)
        Humbert = Stub(Humbert)
        
    def tearDown(self):
        global Humbert
        try:
            Humbert = Humbert.restore()
        except AttributeError:
            pass # already restored ?
    
    def test_replace_as_call(self):
        global Humbert
        def Dostoyevsky():
            return 'DOSTOYEVSKY'
        Humbert.replace( Dostoyevsky )
        
        assert Humbert() is Dostoyevsky()
        assert Humbert() is Dostoyevsky()
        self.assert_humbert_restorable()
        
    def test_replace_as_iter(self):
        global Humbert
        Humbert.replace( [Delores, Lolita] )
    
        assert Humbert() is Delores()
        assert Humbert() is Lolita()
        self.assert_call_humbert_stops()
        self.assert_humbert_restorable()

    def test_replace_as_gen(self):
        global Humbert
        
        def getcalls():
            for fn in (Lolita, Delores):
                yield fn
            
        Humbert.replace(getcalls())
    
        assert Humbert() is Lolita()
        assert Humbert() is Delores()
        self.assert_call_humbert_stops()
        self.assert_humbert_restorable()

def test_mkinterface():
    
    class Complicated(object):
        one = 1
        two = 2
        def __init__(self, three, four):
            self.one = three
            self.two = four
        class sub:
            pass
        class sub2(object):
            pass
        def raise_val_error(self, wants, input):
            raise ValueError
        def raise_attr_error(self):
            raise AttributeError
    
    class LikeCompl(mkinterface(Complicated)):
        pass
    
    assert issubclass(LikeCompl, object), 'unexpected bases: %s' % LikeCompl.__bases__
    n = LikeCompl()
    n.raise_val_error()
    n.raise_attr_error()
    assert n.one == 1, 'unexpected: %s' % n.one
    assert n.two == 2
    assert type(n.sub) == types.ClassType
    assert issubclass(n.sub2, object)
    